% This program performs microaggregation on the GL dataset and plots the
% results shown in Figure 12(c) of [1].
%
% [1]: "Where you are is who you are: User identification by matching statistics",
% Farid M. Naini, Jayakrishnan Unnikrishnan, Patrick Thiran, and Martin
% Vetterli, IEEE Transactions on Information Forensics and Security, 2015



clear all
close all
clc




%% loading the GL data
load GL1


fprintf('Loading the GL data and computing the histograms at 1000 meter resolution.\n');

%% computing the histograms at 1 meter resolution
totData(:,1)=totData(:,1)+1;

% removing users who have only 1 avtive week
numActiveWeeks=zeros(182,1);
for userIter=1:182
    numActiveWeeks(userIter)=length(unique(totData(totData(:,1)==userIter,2)));
end

avtiveUsers = find(numActiveWeeks>=2);
totData(~ismember(totData(:,1),avtiveUsers),:)=[];

% remapping user names to 1-154
[~, ~, newUserMap] = unique(totData(:,1));
totData(:,1)=newUserMap;

% remapping location names
[locIDs, ~, newLocMap] = unique(totData(:,3));
totData(:,3)=newLocMap;

% parameters N and K
N=max(newUserMap);
K=max(newLocMap);

% buffers for the histograms
hist1_1meter=spalloc(N, K, ceil(log(K))*N*2);
hist2_1meter=hist1_1meter;

% iteration on every user
for userIter=1:N
    fprintf('Performing user iteration %d from %d\n', userIter, N);
    curUser_trace = totData(totData(:,1)==userIter,:);
    
    % current user's active weeks
    activeWeekes = unique(curUser_trace(:,2));
    
    %splitting the weeks into two groups
    datasetOneWeeks = activeWeekes(1:ceil(length(activeWeekes)/2));
    datasetTwoWeeks = activeWeekes(ceil(length(activeWeekes)/2)+1:end);
    
    curUser_trace_datasetOne = curUser_trace(ismember(curUser_trace(:,2), datasetOneWeeks),3);
    curUser_trace_datasetTwo = curUser_trace(ismember(curUser_trace(:,2), datasetTwoWeeks),3);
    
    hist1_1meter(userIter,:)=sparse(1, curUser_trace_datasetOne, 1, 1, K);
    hist2_1meter(userIter,:)=sparse(1, curUser_trace_datasetTwo, 1, 1, K);

end


%% computing the histograms at 1000 meter resolution
[hist1_1000meter, hist2_1000meter] = GLHistAggregator(hist1_1meter, hist2_1meter, locIDs, 1000);

hist1=hist1_1000meter;
hist2=hist2_1000meter;



fprintf('\nPerforming microaggregation on the GL histograms:\n');
%% performing matching with no mictoaggregation (i.e, k=1)
hist1MicrAgg = hist1;

% performing the matching
[wMatProp] = weightMatCalcPropMetric(hist1MicrAgg, hist2);

[~, s1Prop, s2Prop] = bipartite_matching(-wMatProp+max(max(wMatProp)));

% computing user accuracy
acProp=sum(s1Prop==s2Prop);


%saving result
userAccRes1=acProp/N*100;
groupAccRes1=userAccRes1;
numClusterRes1=N;
SSERes1=0;
kannLevel1=1;


%% finding the maximum information loss (when k=N)
avrHist1 = mean(hist1,1);
maxL = sum(sum(abs(hist1-repmat(avrHist1, [N,1]))));

%% performing mictoaggregation 

kValues = [1:30,35:5:75];

%buffer for results
userAccRes=zeros(length(kValues),1);
groupAccRes=zeros(length(kValues),1);
numClusterRes=zeros(length(kValues),1);
SSERes=zeros(length(kValues),1);
kannLevel=zeros(length(kValues),1);

% saving the results for k=1
userAccRes(1)=userAccRes1;
groupAccRes(1)=groupAccRes1;
numClusterRes(1)=numClusterRes1;
SSERes(1)=SSERes1/maxL;
kannLevel(1)=kannLevel1;

% performing microaggregation for k>1

for kIter=2:length(kValues)
    
    fprintf('Performing microaggregation iteration %d of %d:\n', kIter, length(kValues));
    
    k=kValues(kIter);
    
    % performing the microaggregation
    [SSE, L, grouping, numC]= GSMS_T2(hist1, k,1);

    hist1MicrAgg = histMicrAgg(hist1, grouping);

    % performing the matching
    [wMatProp] = weightMatCalcPropMetric(hist1MicrAgg, hist2);

    [~, s1Prop, s2Prop] = bipartite_matching(-wMatProp+max(max(wMatProp)));
    
    % computing user accuracy
    acProp=sum(s1Prop==s2Prop);
    
    
    
    % computing group accuracy
    accPropG=groupAcc(s1Prop, s2Prop, grouping);

    %saving result
    userAccRes(kIter)=acProp/N*100;
    groupAccRes(kIter)=accPropG/N*100;
    numClusterRes(kIter)=numC;
    SSERes(kIter)=SSE/maxL*100;
    kannLevel(kIter)=k;
end

%% including the result for k=N
userAccRes(end+1)=1/N*100;
groupAccRes(end+1)=100;
numClusterRes(end+1)=1;
SSERes(end+1)=100;
kannLevel(end+1)=N;


%% plotting the results

figure
semilogx(kannLevel, userAccRes,'bs-','linewidth',2,'markersize',4);
hold on
semilogx(kannLevel, groupAccRes,'rx-','linewidth',2,'markersize',6);
semilogx(kannLevel, SSERes,'g','linewidth',2);
semilogx(kannLevel, numClusterRes/N*100,'k--','linewidth',2);
legend('user-level accuracy', 'cluster-level accuracy', 'information loss', 'normalized # clusters')
xlabel('k');
ylabel('%');
title('Micro-aggregation on the GL dataset');
xlim([1 160])

