% This function loads the GL dataset and computes the histograms at various
% spatial resolutions. It outputs the results given in Section IV-C2 and in
% Figure 11 of [1].
%
% [1]: "Where you are is who you are: User identification by matching statistics",
% Farid M. Naini, Jayakrishnan Unnikrishnan, Patrick Thiran, and Martin
% Vetterli, IEEE Transactions on Information Forensics and Security, 2015




clear all;
close all;
clc;


% loading the histograms
load GL1;


%% computing the histograms at 1 meter resolution
fprintf('Performing preprocessing -- histogram computation\n');
%mapping users names from 0-181 to 1-182
totData(:,1)=totData(:,1)+1;

% removing users who have only 1 avtive week
numActiveWeeks=zeros(182,1);
for userIter=1:182
    numActiveWeeks(userIter)=length(unique(totData(totData(:,1)==userIter,2)));
end

avtiveUsers = find(numActiveWeeks>=2);
totData(~ismember(totData(:,1),avtiveUsers),:)=[];

% remapping user names to 1-154
[~, ~, newUserMap] = unique(totData(:,1));
totData(:,1)=newUserMap;

% remapping location names
[locIDs, ~, newLocMap] = unique(totData(:,3));
totData(:,3)=newLocMap;

% parameters N and K
N=max(newUserMap);
K=max(newLocMap);

% buffers for the histograms
hist1_1meter=spalloc(N, K, ceil(log(K))*N*2);
hist2_1meter=hist1_1meter;

% iteration on every user
for userIter=1:N
    fprintf('Performing user iteration %d from %d\n', userIter, N);
    curUser_trace = totData(totData(:,1)==userIter,:);
    
    % current user's active weeks
    activeWeekes = unique(curUser_trace(:,2));
    
    %splitting the weeks into two groups
    datasetOneWeeks = activeWeekes(1:ceil(length(activeWeekes)/2));
    datasetTwoWeeks = activeWeekes(ceil(length(activeWeekes)/2)+1:end);
    
    curUser_trace_datasetOne = curUser_trace(ismember(curUser_trace(:,2), datasetOneWeeks),3);
    curUser_trace_datasetTwo = curUser_trace(ismember(curUser_trace(:,2), datasetTwoWeeks),3);
    
    hist1_1meter(userIter,:)=sparse(1, curUser_trace_datasetOne, 1, 1, K);
    hist2_1meter(userIter,:)=sparse(1, curUser_trace_datasetTwo, 1, 1, K);

end


%% computing the matching with 1000 meters resolution 
[hist1_1000meter, hist2_1000meter] = GLHistAggregator(hist1_1meter, hist2_1meter, locIDs, 1000);
[numCorMatchProp, numCorMatchL1, numCorMatchCosine, numCorMatchDot]=calculateMatchAllMetrics(hist1_1000meter, hist2_1000meter);
fprintf(['\nHistogram matching for the GL dataset with 1000 meter resolution:\n', ...
        'Proposed metric: %1.2f%% \nl1-norm: %1.2f%%\nCosine distance: %1.2f%%\n', ...
        'Dot product: %1.2f%%\n'], numCorMatchProp/N*100, numCorMatchL1/N*100, ...
        numCorMatchCosine/N*100, numCorMatchDot/N*100);

    
 %% computing the matching with varying resolution   
 fprintf('\nComputing the matching by varying the resolution\n');
 %squareLengthValues=[1 2 3 5 6 10 20 30 50  100 200 300 502 603 1000 1306 1499 2603 2988 3901 6529 7847  13499 19500];   
 squareLengthValues=[ 100 200 300 502 603 1000 1306 1499 2603 2988 3901 6529 7847  13499 19500];   
 % buffer for results
 matchRes=zeros(length(squareLengthValues), 4);
 
 for squareLengthIter = 1:length(squareLengthValues)
     fprintf('Performing gridding iteration %d from %d\n', squareLengthIter, length(squareLengthValues));
      
     [hist1_agr, hist2_agr] = GLHistAggregator(hist1_1meter, hist2_1meter, locIDs, squareLengthValues(squareLengthIter));
     [numCorMatchProp, numCorMatchL1, numCorMatchCosine, numCorMatchDot]=calculateMatchAllMetrics(hist1_agr, hist2_agr);
     
     matchRes(squareLengthIter ,:)=[numCorMatchProp, numCorMatchL1, numCorMatchCosine, numCorMatchDot];
end    

%% plotting the results
figure
semilogx(squareLengthValues, matchRes/N*100, 'linewidth',2);
xlabel('Side lengths of grid squares (meters)');
ylabel('Percentage accuracy');
legend('Proposed metric', 'l_1-norm', 'Cosine distance', 'Dot product', 'Location', 'NorthEast');
title('Matching accuracy on the GL dataset');