% This program performs microaggregation on the WBH dataset and plots the
% results shown in Figure 12(b) of [1].
%
% [1]: "Where you are is who you are: User identification by matching statistics",
% Farid M. Naini, Jayakrishnan Unnikrishnan, Patrick Thiran, and Martin
% Vetterli, IEEE Transactions on Information Forensics and Security, 2015

clear all
close all
clc




%% loading the WBH histograms
load WBH1


%% choosing the subset of users and the subset of locations

locPop = sum((hist1+hist2),1);
[~, locPopInd] = sort(locPop, 'descend');
top5Locs = locPopInd(1:5);

usersSub = (sum(hist1(: , top5Locs),2) > 0) & (sum(hist2(: , top5Locs),2) > 0); 

hist1 = hist1(usersSub, :);
hist2 = hist2(usersSub, :);

K=1000; % considering top 1000 websites

chosenLoc=locPopInd(1:K);
    
% computing the histogram on the chosen subset of websites
hist1Subset=hist1(:, chosenLoc);
hist2Subset=hist2(:, chosenLoc);

% normalizing the histograms
hist1Subset = myHistNorm(hist1Subset);
hist2Subset = myHistNorm(hist2Subset);



hist1 = hist1Subset;
hist2 = hist2Subset;
N=size(hist1,1);    



%% performing matching with no mictoaggregation (i.e, k=1)
hist1MicrAgg = hist1;

% performing the matching
[wMatProp] = weightMatCalcPropMetric(hist1MicrAgg, hist2);

[~, s1Prop, s2Prop] = bipartite_matching(-wMatProp+max(max(wMatProp)));

% computing user accuracy
acProp=sum(s1Prop==s2Prop);


%saving result
userAccRes1=acProp/N*100;
groupAccRes1=userAccRes1;
numClusterRes1=N;
SSERes1=0;
kannLevel1=1;


%% finding the maximum information loss (when k=N)
avrHist1 = mean(hist1,1);
maxL = sum(sum(abs(hist1-repmat(avrHist1, [N,1]))));

%% performing mictoaggregation 

kValues = [1:30,32:2:50];

%buffer for results
userAccRes=zeros(length(kValues),1);
groupAccRes=zeros(length(kValues),1);
numClusterRes=zeros(length(kValues),1);
SSERes=zeros(length(kValues),1);
kannLevel=zeros(length(kValues),1);

% saving the results for k=1
userAccRes(1)=userAccRes1;
groupAccRes(1)=groupAccRes1;
numClusterRes(1)=numClusterRes1;
SSERes(1)=SSERes1/maxL;
kannLevel(1)=kannLevel1;

% performing microaggregation for k>1

for kIter=2:length(kValues)
    fprintf('Performing iteration %d from %d\n', kIter, length(kValues));
    
    k=kValues(kIter);
    
    % performing the microaggregation
    [SSE, L, grouping, numC]= GSMS_T2(hist1, k,1);

    hist1MicrAgg = histMicrAgg(hist1, grouping);

    % performing the matching
    [wMatProp] = weightMatCalcPropMetric(hist1MicrAgg, hist2);

    [~, s1Prop, s2Prop] = bipartite_matching(-wMatProp+max(max(wMatProp)));
    
    % computing user accuracy
    acProp=sum(s1Prop==s2Prop);
    
    
    
    % computing group accuracy
    accPropG=groupAcc(s1Prop, s2Prop, grouping);

    %saving result
    userAccRes(kIter)=acProp/N*100;
    groupAccRes(kIter)=accPropG/N*100;
    numClusterRes(kIter)=numC;
    SSERes(kIter)=SSE/maxL*100;
    kannLevel(kIter)=k;
end

%% including the result for k=N
userAccRes(end+1)=1/N*100;
groupAccRes(end+1)=100;
numClusterRes(end+1)=1;
SSERes(end+1)=100;
kannLevel(end+1)=N;

%% plotting the results

figure
semilogx(kannLevel, userAccRes,'bs-','linewidth',2,'markersize',4);
hold on
semilogx(kannLevel, groupAccRes,'rx-','linewidth',2,'markersize',6);
semilogx(kannLevel, SSERes,'g','linewidth',2);
semilogx(kannLevel, numClusterRes/N*100,'k--','linewidth',2);
legend('user-level accuracy', 'cluster-level accuracy', 'information loss', 'normalized # clusters')
xlabel('k');
ylabel('%');
title('Micro-aggregation on the WBH dataset');
xlim([1 110])


